from eptr2 import EPTR2
import pandas as pd
from eptr2.util.time import iso_to_contract


def get_day_ahead_and_bilateral_matches(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    verbose: bool = False,
    include_contract_symbol: bool = False,
    org_id: str = None,
):
    """
    This composite function gets day ahead and bilateral matches for the whole market or for a single organization.
    """

    if verbose:
        print("Getting day ahead matches...")

    df_da = eptr.call(
        "dam-clearing", start_date=start_date, end_date=end_date, org_id=org_id
    )

    df = df_da.rename(
        {"matchedBids": "da_long", "matchedOffers": "da_short"}, axis=1
    ).drop("hour", axis=1)

    if verbose:
        print("Getting bilateral long matches...")

    df_bi_long = eptr.call(
        "bi-long", start_date=start_date, end_date=end_date, org_id=org_id
    )

    df = df.merge(
        df_bi_long.rename({"quantity": "bi_long"}, axis=1).drop("hour", axis=1),
        on="date",
        how="left",
    )

    if verbose:
        print("Getting bilateral short matches...")

    df_bi_short = eptr.call(
        "bi-short", start_date=start_date, end_date=end_date, org_id=org_id
    )

    df = df.merge(
        df_bi_short.rename({"quantity": "bi_short"}, axis=1).drop("hour", axis=1),
        on="date",
        how="left",
    )

    if include_contract_symbol:
        try:
            df["contract"] = df["date"].apply(lambda x: iso_to_contract(x))
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    return df


def get_day_ahead_detail_info(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    verbose: bool = False,
    include_contract_symbol: bool = False,
    lives: int = 3,
    **kwargs,
) -> pd.DataFrame:
    """
    This function gets day ahead detail information such as regular, block, flexible bids & asks (offers).
    """

    df_d: dict[pd.DataFrame] = {}

    items = [
        "dam-bid",
        "dam-offer",
        "dam-block-bid",
        "dam-block-offer",
        "dam-flexible-bid",
        "dam-flexible-offer",
        # "dam-flexible-matching",
    ]

    map_d = {
        "dam-bid": {"bidQuantity": "bid"},
        "dam-offer": {"offerQuantity": "ask"},
        "dam-block-bid": {
            "amountOfPurchasingTowardsMatchedBlock": "block_long_match",
            "amountOfPurchasingTowardsUnMatchedBlock": "block_long_remaining",
        },
        "dam-block-offer": {
            "amountOfSalesTowardsMatchedBlock": "block_short_match",
            "amountOfSalesTowardsUnMatchedBlock": "block_short_remaining",
        },
        "dam-flexible-bid": {
            "totalBuyingFlexibleOfferQuantity": "flex_long_total",
            "matchedBuyingFlexibleOfferQuantity": "flex_long_match",
            "unmatchedBuyingFlexibleOfferQuantity": "flex_long_remaining",
        },
        "dam-flexible-offer": {
            "totalSellingFlexibleOfferQuantity": "flex_short_total",
            "matchedSellingFlexibleOfferQuantity": "flex_short_match",
            "unmatchedSellingFlexibleOfferQuantity": "flex_short_remaining",
        },
    }

    df: pd.DataFrame | None = None

    while len(items) > 0:
        for item in items:
            try:
                if verbose:
                    print(f"Fetching {item} data...")
                df_d[item] = eptr.call(
                    item,
                    start_date=start_date,
                    end_date=end_date,
                    request_kwargs={"timeout": kwargs.get("timeout", 5)},
                )

                df_d[item].rename(map_d[item], axis=1, inplace=True)
                if "hour" in df_d[item].columns:
                    df_d[item].drop("hour", axis=1, inplace=True)
                elif "time" in df_d[item].columns:
                    df_d[item].drop("time", axis=1, inplace=True)

                if df is None:
                    df = df_d[item].copy()
                else:
                    df = df.merge(df_d[item], on="date", how="outer")

            except Exception as e:
                if verbose:
                    print(f"Error fetching {item}: {e}")
                lives -= 1
                if lives <= 0:
                    print("Max lives reached. Exiting.")
                    break
                continue

        items = [item for item in items if item not in df_d or df_d[item].empty]

    # df.fillna(0, inplace=True)

    if include_contract_symbol:
        try:
            df["contract"] = df["date"].apply(lambda x: iso_to_contract(x))
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    return df



from typing import Any
import urllib3
import re
import os
import json
from urllib.parse import urljoin
import copy
from eptr2.mapping import (
    get_total_path,
    get_call_method,
    get_required_parameters,
    get_param_label,
    get_path_map,
    get_optional_parameters,
    alias_to_path,
    get_alias_map,
    get_derived_calls,
)
from warnings import warn
from eptr2.processing.preprocess import preprocess_parameter, process_special_calls
from datetime import datetime, timedelta


class EPTR2:
    def __init__(
        self,
        username: str = None,
        password: str = None,
        tgt_d: dict | None = None,
        credentials_file_path: str | None = None,
        **kwargs,
    ) -> None:
        ## kwargs are
        ### map_param_labels: bool
        ### secure: bool
        ### query_parameters: dict
        ### just_call_phrase: bool
        self.ssl_verify = kwargs.get("ssl_verify", True)
        self.check_postprocess(postprocess=kwargs.get("postprocess", True))
        self.get_raw_response = kwargs.get("get_raw_response", False)

        ### Credentials and Login
        self.username = username
        self.password = password
        self.is_test = kwargs.get("is_test", False)  ## Currently not used
        self.credentials_file_path = credentials_file_path
        # self.credentials_file_path = kwargs.get("credentials_file_path", None)
        self.login(custom_root_phrase=kwargs.get("root_phrase", None))

        if tgt_d is not None:
            self.import_tgt_info(tgt_d)
        else:
            self.tgt = None
            self.tgt_exp = 0
            self.tgt_exp_0 = 0

        self.check_renew_tgt()

        ## Path map keys and custom aliases
        self.path_map_keys = get_path_map(just_call_keys=True)
        self.custom_aliases = kwargs.get("custom_aliases", {})

    def login(self, custom_root_phrase: str | None = None):
        if self.username is None or self.password is None:
            if self.credentials_file_path is not None:
                with open(self.credentials_file_path, "r") as f:
                    credentials_d = json.load(f)
                    self.username = credentials_d["EPTR_USERNAME"]
                    self.password = credentials_d["EPTR_PASSWORD"]
            else:
                self.username = os.environ.get("EPTR_USERNAME", None)
                self.password = os.environ.get("EPTR_PASSWORD", None)

        if self.username is None or self.password is None:
            raise Exception(
                "Username and password must be provided for login. If you do not have the necessary credentials, you can get them from EPIAS Transparency Platform website."
            )

        if not custom_root_phrase:
            root_phrase_test = "-prp" if self.is_test else ""
            root_phrase_default = f"https://seffaflik{root_phrase_test}.epias.com.tr"
            self.root_phrase = root_phrase_default

    ## Ref: https://stackoverflow.com/a/62303969/3608936
    def __getattr__(self, __name: str) -> Any:

        def method(*args, **kwargs):
            key_raw = __name
            key = re.sub("_", "-", key_raw)
            key = alias_to_path(alias=key, custom_aliases=self.custom_aliases)
            if key not in self.path_map_keys:
                raise Exception(
                    "This call is not yet defined. Call 'get_available_calls' method to see the available calls."
                )
            else:
                warn("This method is deprecated. Use 'call' method instead.")
            # required_body_params = get_required_parameters(key)
            return getattr(self, "call")(key=key, **kwargs)

        return method

    def import_tgt_info(self, tgt_d):
        self.tgt = tgt_d["tgt"]
        self.tgt_exp = tgt_d["tgt_exp"]
        self.tgt_exp_0 = tgt_d["tgt_exp_0"]

    def check_renew_tgt(self):
        if self.tgt is None or self.tgt_exp_0 < datetime.now().timestamp():
            self.get_tgt()

    def get_tgt(self, **kwargs):
        if self.username is None or self.password is None:
            raise Exception("Username and password must be provided for tgt renewal.")

        test_suffix = "-prp" if self.is_test else ""
        login_url = f"""https://giris{test_suffix}.epias.com.tr/cas/v1/tickets"""
        login_url += f"?username={self.username}&password={self.password}"

        http = urllib3.PoolManager(
            cert_reqs="CERT_REQUIRED" if self.ssl_verify else "CERT_NONE"
        )

        res = http.request(
            method="POST",
            url=login_url,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
            },
            **kwargs.get("request_kwargs", {}),
        )
        if res.status not in [200, 201]:
            raise Exception(
                "Request failed with status code: "
                + str(res.status)
                + "\n"
                + res.data.decode("utf-8")
            )

        res_data = json.loads(res.data.decode("utf-8"))
        self.tgt = res_data["tgt"]
        try:
            tgt_start_time = datetime.fromisoformat(res_data["created"])
        except:
            tgt_start_time = datetime.now()

        ## Hard timeout
        self.tgt_exp = (tgt_start_time + timedelta(hours=1, minutes=45)).timestamp()

        ## Soft timeout
        self.tgt_exp_0 = min(
            self.tgt_exp,
            (tgt_start_time + timedelta(hours=1, minutes=45)).timestamp(),
        )

    def export_tgt_info(self):
        """
        Exports TGT information to a dictionary. Contents are TGT itself, expiration timestamp (tgt_exp) and soft expiration (tgt_exp_0 i.e. expiration if not used) timestamp.
        """
        return {
            "tgt": self.tgt,
            "tgt_exp": self.tgt_exp,
            "tgt_exp_0": self.tgt_exp_0,
        }

    def check_postprocess(self, postprocess: bool = True):
        self.postprocess = postprocess
        if self.postprocess:
            try:
                from eptr2.mapping.processing import get_postprocess_function
            except ImportError:
                print(
                    "pandas is not installed. Some functionalities may not work properly. Postprocessing is disabled. To disable postprocessing just set 'postprocess' parameter to False when calling EPTR2 class.",
                )
                self.postprocess = False

    def get_available_calls(self, include_aliases: bool = False):
        """
        Gets all the available calls of eptr2 package. As a reminder, number of calls at eptr2 package might be lower than the actual calls. If include_aliases is set to True, it also includes the aliases.
        """

        if include_aliases:

            return {
                "keys": self.path_map_keys,
                "default_aliases": get_alias_map(),
                "custom_aliases": self.custom_aliases,
            }

        return self.path_map_keys

    def get_number_of_calls(self):
        """
        List the number of calls in the package. It also lists the number of derived calls and API calls (excluding derived calls).
        """

        d = {
            "all_calls": self.path_map_keys,
            "derived_calls": get_derived_calls(),
        }
        d["n_total_calls"] = len(d["all_calls"])
        d["n_derived_calls"] = len(d["derived_calls"])
        d["n_api_calls"] = len(d["all_calls"]) - len(d["derived_calls"])

        return d

    def get_aliases(self, include_custom_aliases: bool = False):
        """
        Gets only the aliases. If include_custom_aliases is set to True, it also includes the user defined aliases (custom_aliases).
        """

        alias_d = get_alias_map()
        if include_custom_aliases:
            alias_d.update(self.custom_aliases)

        return alias_d

    def call(self, key: str, **kwargs):
        """
        Main call function for the API. This function is used to process parameters and make calls to EPIAS Transparency API.
        """

        self.check_renew_tgt()
        raw_key = key
        key = alias_to_path(alias=key, custom_aliases=self.custom_aliases)

        if key not in self.path_map_keys:
            if raw_key == key:
                raise Exception(
                    f"This call {raw_key} is not yet defined in calls or aliases. Call 'get_available_calls' method to see the available calls."
                )
            else:
                raise Exception(
                    f"This alias {raw_key} forwarded to key {key} is not yet defined in calls. Call 'get_available_calls' method to see the available calls."
                )

        call_path = get_total_path(key)
        call_method = get_call_method(key)
        required_body_params = get_required_parameters(key)
        call_body_raw = kwargs.pop("call_body", kwargs)

        ### There are some calls requiring special handling, they have a special function to process them
        call_body_raw = process_special_calls(key, call_body_raw)

        optional_body_params = get_optional_parameters(key)
        all_params = required_body_params + optional_body_params

        call_body = {
            k: preprocess_parameter(k, v)
            for k, v in call_body_raw.items()  ## If there is a call_body parameter in kwargsi, use it, else use kwargs
            if k in all_params
        }

        for body_key in required_body_params:
            if body_key not in call_body.keys():
                call_body[body_key] = preprocess_parameter(body_key, None)
            kwargs.pop(body_key, None)

        if call_body is not None:
            if not all([x in call_body.keys() for x in required_body_params]):
                raise Exception("Some required parameters are missing in call body.")

            if kwargs.get("map_param_labels", True):
                cb2 = {}
                for k, v in call_body.items():
                    ## Updated this part to handle multiple labels originating from a single label
                    label = get_param_label(k)["label"]
                    value = v
                    if isinstance(label, list):
                        for l in label:
                            cb2[l] = value
                    else:
                        cb2[label] = value
                call_body = copy.deepcopy(cb2)

        elif len(required_body_params) > 0:
            raise Exception("Required parameters are missing in call body.")

        ## If the call is uevm, change powerPlantId to powerplantId due to only non-standard naming in powerPlantId
        if key == "uevm" and "powerPlantId" in call_body.keys():
            call_body["powerplantId"] = call_body.pop("powerPlantId")

        res = transparency_call(
            call_path=call_path,
            call_method=call_method,
            call_body=call_body,
            root_phrase=self.root_phrase,
            ssl_verify=self.ssl_verify,
            is_test=self.is_test,
            tgt=self.tgt,
            **kwargs,
        )

        ## Set soft timeout for tgt renewal
        self.tgt_exp_0 = min(
            self.tgt_exp,
            datetime.now().timestamp() + 60 * 90,
        )

        if kwargs.get("get_raw_response", self.get_raw_response):
            return res

        res = json.loads(res.data.decode("utf-8"))
        if kwargs.get("postprocess", self.postprocess):
            from eptr2.mapping.processing import get_postprocess_function

            df = get_postprocess_function(key)(res, key=key)
            return df

        return res


def transparency_call(
    call_path: dict,
    call_method: str,
    call_body: dict | None = None,
    is_test: bool = False,
    tgt: str = None,
    **kwargs,
):
    """
    Core function to perform the calls and return results. This function is used by the EPTR2 class to make calls to EPIAS Transparency API.
    """

    ## kwargs are
    ### secure: bool
    ### query_parameters: dict
    ### just_call_phrase: bool

    root_phrase_test = "-prp" if is_test else ""
    root_phrase_default = f"https://seffaflik{root_phrase_test}.epias.com.tr"
    root_phrase = kwargs.get("root_phrase", root_phrase_default)

    call_phrase = urljoin(root_phrase, call_path)

    ## If there are query parameters (not body parameters), append
    q_params = kwargs.get("query_parameters", {})

    if len(q_params.keys()) > 0:
        call_phrase += "?" + "&".join(
            [str(k) + "=" + str(v) for k, v in q_params.items()]
        )

    if kwargs.pop("just_call_phrase", False):
        return call_phrase

    if call_body is not None and call_body != {} and call_method == "GET":
        raise Exception("GET method does not allow body parameters.")

    ssl_verify = kwargs.pop("ssl_verify", True)

    http = urllib3.PoolManager(cert_reqs="CERT_REQUIRED" if ssl_verify else "CERT_NONE")

    header_d = {"Content-Type": "application/json"}
    if tgt is not None:
        header_d["TGT"] = tgt

    full_url = urljoin(call_phrase, "")
    res = http.request(
        method=call_method,
        url=full_url,
        body=json.dumps(call_body),
        headers=header_d,
        **kwargs.get("request_kwargs", {}),
    )

    if res.status not in [200, 201]:
        raise Exception(
            "Request failed with status code: "
            + str(res.status)
            + "\n"
            + res.data.decode("utf-8")
        )
    return res


def set_eptr_credentials_to_env(cred_path: str = "creds/eptr_credentials.json"):
    """Set EPTR credentials to environment variables from a JSON file."""

    if not os.path.exists(cred_path):
        raise FileNotFoundError(
            f"Credentials file not found at {cred_path}. Please provide a valid path or put your eptr_credentials.json file there."
        )

    try:
        with open(cred_path, "r") as f:
            cred_d: dict = json.load(f)

        os.environ["EPTR_USERNAME"] = cred_d["EPTR_USERNAME"]
        os.environ["EPTR_PASSWORD"] = cred_d["EPTR_PASSWORD"]
        if "tgt_d" in cred_d.keys():
            os.environ["tgt_d"] = json.dumps(cred_d["tgt_d"])
    except Exception as e:
        raise ValueError(
            f"Error while reading the credentials file from path {cred_path}. Please check the file format and content. Error: {e}"
        ) from e


def eptr_w_tgt_wrapper(
    cred_path: str = "creds/eptr_credentials.json", **kwargs
) -> EPTR2:
    """This function is a wrapper for the EPTR2 class to handle TGT (Ticket Granting Ticket) management and credentials loading. This way TGT is automatically renewed when it expires, and credentials are loaded from a file or environment variables."""

    ### Initially you can define EPTR_USERNAME and EPTR_PASSWORD in the keyword arguments (strictly optional) and even tgt_d optional parameter. It is not a conventional use of this function.
    username = kwargs.get("EPTR_USERNAME", None)
    password = kwargs.get("EPTR_PASSWORD", None)
    tgt_d = None

    ### If username and password are not defined in the keyword arguments, try to get them from the environment variables.
    if username is not None:
        os.environ["EPTR_USERNAME"] = username
    else:
        username = os.getenv("EPTR_USERNAME", None)
    if password is not None:
        os.environ["EPTR_PASSWORD"] = password
    else:
        password = os.getenv("EPTR_PASSWORD", None)

    ### Check for EPTR_USERNAME and EPTR_PASSWORD in the environment variables. If they are not set, try to get them from the credentials file.
    if any([k is None for k in [username, password]]):
        set_eptr_credentials_to_env(cred_path)
        username = os.getenv("EPTR_USERNAME", None)
        password = os.getenv("EPTR_PASSWORD", None)
        tgt_d = os.getenv("tgt_d", None)
        if tgt_d is not None:
            tgt_d = json.loads(tgt_d)
    else:
        tgt_d = kwargs.get("tgt_d", None)
        if tgt_d is not None:
            os.environ["tgt_d"] = json.dumps(tgt_d)

        if tgt_d is None:
            tgt_d = os.getenv("tgt_d", None)
            if tgt_d is not None:
                tgt_d = json.loads(tgt_d)

    ### Check for EPTR_USERNAME and EPTR_PASSWORD in the environment variables. If they are not set, raise an error. (Sanity check)

    if any([k is None for k in [username, password]]):
        raise ValueError(
            "EPTR_USERNAME and/or EPTR_PASSWORD is not set in the environment variables. They are also not acquired from the credentials file."
        )

    eptr = EPTR2(
        username=username,
        password=password,
        tgt_d=tgt_d,
    )

    if kwargs.get("skip_tgt_update", False):
        return eptr

    ### Export the tgt information and check with the existing one. If they are different, update the environment variables and the credentials file.
    tgt_d2 = eptr.export_tgt_info()

    if tgt_d != tgt_d2:
        os.environ["tgt_d"] = json.dumps(tgt_d2)

        if all([x in os.environ.keys() for x in ["EPTR_USERNAME", "EPTR_PASSWORD"]]):
            with open(cred_path, "w") as f:
                json.dump(
                    {
                        "EPTR_USERNAME": os.getenv("EPTR_USERNAME", None),
                        "EPTR_PASSWORD": os.getenv("EPTR_PASSWORD", None),
                        "tgt_d": tgt_d2,
                    },
                    f,
                )
        else:
            print(
                "EPTR_USERNAME and/or EPTR_PASSWORD are not set in the environment variables. Credentials file is not updated."
            )

    return eptr


def generate_eptr2_credentials_file(
    username: str = "YOUR_USERNAME",
    password: str = "YOUR_PASSWORD",
    cred_path: str = "creds/eptr_credentials.json",
    overwrite: bool = False,
):
    """Generate a credentials file for EPTR2 class. It creates a JSON file with EPTR_USERNAME and EPTR_PASSWORD."""

    if not overwrite and os.path.exists(cred_path):
        print(
            f"Credentials file already exists at {cred_path}. Use overwrite=True to overwrite."
        )
        return

    if not os.path.exists(os.path.dirname(cred_path)):
        os.makedirs(os.path.dirname(cred_path))

    cred_d = {
        "EPTR_USERNAME": username,
        "EPTR_PASSWORD": password,
    }

    with open(cred_path, "w") as f:
        json.dump(cred_d, f)

    print(f"Credentials file created at {cred_path}.")



from eptr2 import EPTR2
import pandas as pd
from eptr2.util.time import (
    iso_to_contract,
    datetime_to_contract,
    check_date_for_settlement,
)


def get_hourly_production_data(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    rt_pp_id: str | int | None = None,
    uevm_pp_id: str | int | None = None,
    verbose: bool = False,
    include_contract_symbol: bool = False,
):
    """
    This composite function gets production data (Gerçek Zamanlı Üretim, UEVM) and merges them.

    It is also possible to enter pp_id to get the production data for a specific production plan. Example ID values are given below.

    rt_pp_id=641, ## ATATÜRK HES
    uevm_pp_id=142, ## ATATÜRK HES
    """

    if verbose:
        print("Loading real time production data...")

    rt_gen_df: pd.DataFrame = eptr.call(
        "rt-gen",
        start_date=start_date,
        end_date=end_date,
        pp_id=rt_pp_id,
    )

    if rt_gen_df.empty:
        raise ValueError("No data (production) is available for this date range.")

    try:
        rt_gen_df.drop("hour", axis=1, inplace=True)
    except Exception as e:
        if verbose:
            print(e)

    rt_gen_df.columns = [
        x + "_rt" if x not in ["date"] else x for x in rt_gen_df.columns
    ]

    if verbose:
        print("Loading UEVM data...")

    uevm_df: pd.DataFrame = eptr.call(
        "uevm",
        start_date=start_date,
        end_date=end_date,
        pp_id=uevm_pp_id,
    )

    try:
        rt_gen_df.drop("hour", axis=1, inplace=True)
    except Exception as e:
        within_settlement = check_date_for_settlement(x=end_date)
        if not within_settlement:
            print("Warning: The end date may not be within the settlement period.")
        if verbose:
            print(e)

    uevm_df.columns = [x + "_uevm" if x not in ["date"] else x for x in uevm_df.columns]

    if verbose:
        print("Merging data...")

    if uevm_df.empty:
        merged_df = rt_gen_df.copy()
    else:
        merged_df = rt_gen_df.merge(uevm_df, how="left", on=["date"])

    if include_contract_symbol:
        try:
            merged_df["contract"] = merged_df["date"].apply(
                lambda x: iso_to_contract(x)
            )
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    merged_df = merged_df.rename(columns={"date": "dt"})

    return merged_df


def get_hourly_production_plan_data(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    org_id: str | None = None,
    uevcb_id: str | None = None,
    verbose: bool = False,
    include_contract_symbol: bool = False,
):
    """
    This composite function gets KGUP v1, KGUP and KUDUP data and merges them.

    It is also possible to enter org_id to get the total production plan for a specific organization. If uevcb_id is also added with org_id, it will get the production plan for a specific uevcb. Example ID values are given below.

    org_id=195, ## ELEKTRİK ÜRETİM AŞ
    uevcb_id=733, ## ATATÜRK HES DB
    """

    if uevcb_id is not None:
        if org_id is None:
            raise ValueError("org_id is required if uevcb_id is specified.")

    if verbose:
        print("Loading KGUP v1...")

    kgupv1_df: pd.DataFrame = eptr.call(
        "kgup-v1",
        start_date=start_date,
        end_date=end_date,
        org_id=org_id,
        uevcb_id=uevcb_id,
    )

    if kgupv1_df.empty:
        raise ValueError("No data (KGUP v1) is available for this date range.")

    kgupv1_df.columns = [
        x + "_kgup_v1" if x not in ["date", "time"] else x for x in kgupv1_df.columns
    ]

    if verbose:
        print("Loading KGUP...")

    kgup_df: pd.DataFrame = eptr.call(
        "kgup",
        start_date=start_date,
        end_date=end_date,
        org_id=org_id,
        uevcb_id=uevcb_id,
    )

    kgup_df.columns = [
        x + "_kgup" if x not in ["date", "time"] else x for x in kgup_df.columns
    ]

    if verbose:
        print("Loading KUDUP...")

    kudup_df: pd.DataFrame = eptr.call(
        "kudup",
        start_date=start_date,
        end_date=end_date,
        org_id=org_id,
        uevcb_id=uevcb_id,
    )

    kudup_df.columns = [
        x + "_kudup" if x not in ["date", "time"] else x for x in kudup_df.columns
    ]

    if verbose:
        print("Merging dataframes...")

    merged_df = kgupv1_df.merge(kgup_df, on=["date", "time"], how="outer")
    merged_df = merged_df.merge(kudup_df, on=["date", "time"], how="outer")

    first_cols = ["date", "time"] + [
        x for x in merged_df.columns if x.startswith("toplam")
    ]

    merged_df = merged_df[
        first_cols + [x for x in merged_df.columns if x not in first_cols]
    ]

    if include_contract_symbol:
        try:
            merged_df["contract"] = merged_df["date"].apply(
                lambda x: iso_to_contract(x)
            )
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    merged_df = merged_df.rename(columns={"date": "dt"})

    col_order = ["dt"] + ["contract"] if include_contract_symbol else []

    merged_df = merged_df[
        col_order + [x for x in merged_df.columns if x not in col_order]
    ]

    return merged_df


def wrapper_hourly_production_plan_and_realized(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    org_id: str | None = None,
    uevcb_id: str | None = None,
    rt_pp_id: str | None = None,
    uevm_pp_id: str | None = None,
    verbose: bool = False,
    include_contract_symbol: bool = False,
):

    if verbose:
        print("Loading production plan data...")

    plan_df = get_hourly_production_plan_data(
        eptr=eptr,
        start_date=start_date,
        end_date=end_date,
        org_id=org_id,
        uevcb_id=uevcb_id,
        verbose=verbose,
        include_contract_symbol=include_contract_symbol,
    )

    if verbose:
        print("Loading production realizations data...")

    realized_df = get_hourly_production_data(
        eptr=eptr,
        start_date=start_date,
        end_date=end_date,
        rt_pp_id=rt_pp_id,
        uevm_pp_id=uevm_pp_id,
        verbose=verbose,
        include_contract_symbol=False,
    )

    merged_df = plan_df.merge(realized_df, how="outer", on=["dt"])

    return merged_df



def calculate_error_list(
    actual_values: list, forecast_values: list, type: str, return_average=False
):
    if type == "ae":
        error_list = [abs(x - y) for x, y in zip(forecast_values, actual_values)]
    elif type == "bias":
        error_list = [(x - y) for x, y in zip(forecast_values, actual_values)]
    elif type == "pbias":
        error_list = [(x - y) / y for x, y in zip(forecast_values, actual_values)]
    elif type == "ape":
        error_list = [abs((x - y) / y) for x, y in zip(forecast_values, actual_values)]
    elif type == "se":
        error_list = [(x - y) ** 2 for x, y in zip(forecast_values, actual_values)]
    else:
        raise ValueError("type must be one of ae, bias, pbias or ape")

    if return_average:
        return sum(error_list) / len(error_list)

    return error_list


def calculate_wmape(actual_values: list, forecast_values: list):
    if len(actual_values) != len(forecast_values):
        raise ValueError("Actual and forecast values must have the same length")

    wmape = sum(
        calculate_error_list(
            actual_values=actual_values, forecast_values=forecast_values, type="ae"
        )
    ) / sum(actual_values)

    return wmape



from eptr2.processing.postprocess.items import (
    postprocess_items_to_df,
    postprocess_mcp_status,
    postprocess_direct_dict,
    postprocess_direct_dict_to_df,
    postprocess_ren_capacity_dict_to_df,
)


def get_postprocess_function(key):
    if key in [
        "dam-volume",
        "pi-offer",
        "pi-bid",
        "supply-demand",
        "dam-bid",
        "dam-offer",
        "dam-block-bid",
        "dam-block-offer",
        "dam-flexible-bid",
        "dam-flexible-offer",
        "dam-flexible-matching",
        "dam-clearing",
        "dam-clearing-org-list",
        "dam-diff",
        "smp",
        "smp-dir",
        "bpm-up",
        "bpm-down",
        "mcp",
        "wap",
        "idm-qty",
        "idm-mm-bid",
        "idm-mm-offer",
        "idm-mm-matching",
        "idm-volume",
        "idm-log",
        "idm-ob-qty",
        "bi-long",
        "bi-short",
        "bi-euas",
        "imb-qty",
        "imb-vol",
        "imb-qty-g",
        "imb-org-list",
        "interim-mcp",
        "mcp-smp-imb",
        "bpm-orders-w-avg",
        "bpm-orders",
        "market-participants",
        "market-participants-organization-list",
        "participant-count-based-upon-license-type",
        "kgup",
        "kgup-v1",
        "kudup",
        "eak",
        "gen-org",
        "gen-uevcb",
        "rt-gen",
        "pp-list",
        "uevm-pp-list",
        "lic-pp-list",
        "load-plan",
        "rt-cons",
        "uecm",
        "st-uecm",
        "su-uecm",
        "uevm",
        "wind-forecast",
        "ren-rt-gen",
        "ren-ul-gen",
        "ren-ul-cost",
        "ren-unit-cost",
        "ren-income",
        "ren-total-cost",
        "ren-participant-list",
        "ren-uevm",
        "ren-lic-cost",
        "zero-balance",
        "iskk",
        "congestion-cost",
        "eic-x-org-list",
        "eic-w-org-list",
        "eic-w-uevcb-list",
        "mms",
        "mms-pp-list",
        "region-list",
        "mms-message-type-list",
        "mms-region-list",
        "mms-uevcb-list",
        "tcat-pre-year-forecast",
        "tcat-pre-month-forecast",
        "international-line-events",
        "line-capacities",
        "capacity-demand",
        "nominal-capacity",
        "dams-active-fullness",
        "dams-daily-level",
        "dams-active-volume",
        "dams-daily-volume",
        "dams-level-minmax",
        "dams-volume-minmax",
        "dams-info",
        "dams-water-energy-provision",
        "idm-summary",
        "electricity-market-quantity",
        "idm-contract-list",
        "idm-order-history",
        "anc-pf-qty",
        "anc-sf-qty",
        "anc-pfk",
        "anc-sfk",
        "long-term-demand-forecast",
        "consumption-breakdown",
        "consumer-breakdown",
        "distribution-region-list",
        "ra-distribution-list",
        "ra-organization-list",
        "ra-spg-list",
        "ra-vspg-list",
        "ra-meters",
        "ra-meter-volumes-period",
        "ra-meter-volumes-version",
        "ra-sum",
        "ng-participants",
        "ng-participant-list",
        "ng-balancing-notifications",
        "ng-balancing-price",
        "ng-bast",
        "ng-blue-code-ops",
        "ng-daily-match-qty",
        "ng-drp",
        "ng-daily-trade-volume",
        "ng-code-four-ops",
        "ng-gddk",
        "ng-green-code-ops",
        "ng-grp-match-qty",
        "ng-grp-trade-volume",
        "ng-imbalance-amount",
        "ng-imbalance-system",
        "ng-match-quantity",
        "ng-orange-code-ops",
        "ng-physical-realization",
        "ng-spot-prices",
        "ng-shippers-imbalance-quantity",
        "ng-system-direction",
        "ng-total-trade-volume",
        "ng-transaction-history",
        "ng-virtual-realization",
        "ng-weekly-matched-quantity",
        "ng-wrp",
        "ng-weekly-trade-volume",
        "ng-vgp-contract-price-summary",
        "ng-vgp-contract-price-summary-period",
        "ng-vgp-contract-price-summary-se",
        "ng-vgp-delivery-period",
        "ng-vgp-delivery-year",
        "ng-vgp-ggf",
        "ng-vgp-ggf-period",
        "ng-vgp-ggf-se",
        "ng-vgp-matched-quantity",
        "ng-vgp-matched-quantity-period",
        "ng-vgp-matched-quantity-se",
        "ng-vgp-open-positions",
        "ng-vgp-open-positions-period",
        "ng-vgp-open-positions-se",
        "ng-vgp-order-book",
        "ng-vgp-transaction-history",
        "ng-vgp-transaction-history-period",
        "ng-vgp-transaction-history-se",
        "ng-vgp-transaction-volumes",
        "ng-vgp-transaction-volumes-period",
        "ng-vgp-transaction-volumes-se",
        "ng-tr-capacity-point",
        "ng-tr-daily-transmission",
        "ng-tr-day-ahead",
        "ng-tr-day-end",
        "ng-tr-entry-nomination",
        "ng-tr-exit-nomination",
        "ng-tr-max-entry-amount",
        "ng-tr-max-exit-amount",
        "ng-tr-actual-entry-amount",
        "ng-tr-actual-exit-amount",
        "ng-tr-reserved-entry-amount",
        "ng-tr-reserved-exit-amount",
        "ng-tr-stock-amount",
        "ng-tr-storage-facility-list",
        "ng-tr-bilateral-transfer",
        "menu",
        "yek-g-bilateral-contracts",
        "yek-g-redemptions",
        "yek-g-expirations",
        "yek-g-issued",
        "yek-g-match-price-minmax",
        "yek-g-order-quantity",
        "yek-g-trade-volume",
        "yek-g-wap",
        "yek-g-withdrawals",
        "yek-g-match-quantity",
        "vep-contract-price-summary",
        "vep-delivery-period-list",
        "vep-delivery-year-list",
        "vep-ggf",
        "vep-ggf-period",
        "vep-load-types",
        "vep-price-summaries",
        "vep-open-positions",
        "vep-trade-volume",
        "vep-transaction-history",
        "vep-transaction-history-periods",
        "vep-matching-quantity",
        "eligible-consumer-count-detail",
        "eligible-consumer-quantity",
        "monthly-index",
        "multiple-factor",
        "percentage-consumption-info",
        "planned-outages",
        "eligible-consumer-count",
        "unplanned-outages",
    ]:
        return postprocess_items_to_df

    elif key in ["interim-mcp-status"]:
        return postprocess_mcp_status

    elif key in [
        "date-init",
        "intl-direction-list",
        "intl-capacity-demand-direction-list",
        "basin-list",
        "dam-list",
        "ng-latest-settlement-date",
        "page-settings",
    ]:
        return postprocess_direct_dict

    elif key in [
        "profile-group-list",
        "province-list",
        "district-list",
        "ren-pp-list",
        "get-distribution-companies",
        "main-tariff-group-list",
        "mf-distribution",
        "mf-meter-reading-type",
        "mf-profile-group",
        "elig-profile-groups",
    ]:
        return postprocess_direct_dict_to_df

    elif key in ["ren-capacity"]:
        return postprocess_ren_capacity_dict_to_df
    else:
        raise Exception("Postprocess function is not defined for this call.")



from eptr2 import EPTR2
import pandas as pd
from eptr2.util.time import iso_to_contract


def get_hourly_consumption_and_forecast_data(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    verbose: bool = False,
    include_contract_symbol: bool = False,
):
    """
    This composite function gets load plan, UECM (settlement consumption), real time and consumption data. If end date is after the last settlement data, UECM is filled with real time consumption under consumption column.
    """

    if verbose:
        print("Loading load plan...")

    lp_df = eptr.call(
        "load-plan",
        start_date=start_date,
        end_date=end_date,
        request_kwargs={"timeout": 5},
    )

    df = lp_df[["date", "lep"]].rename(columns={"lep": "load_plan", "date": "dt"})

    if verbose:
        print("Loading UECM...")

    uecm_df: pd.DataFrame = eptr.call(
        "uecm", start_date=start_date, end_date=end_date, request_kwargs={"timeout": 5}
    )

    if not uecm_df.empty:
        uecm_df = uecm_df[["period", "swv"]].rename(
            columns={"period": "dt", "swv": "uecm"}
        )

        df = df.merge(
            uecm_df,
            on="dt",
            how="outer",
        )

    if verbose:
        print("Loading real time consumption...")

    rt_cons = eptr.call(
        "rt-cons",
        start_date=start_date,
        end_date=end_date,
        request_kwargs={"timeout": 5},
    )

    df = df.merge(
        rt_cons[["date", "consumption"]].rename(
            columns={"date": "dt", "consumption": "rt_cons"}
        ),
        on="dt",
        how="outer",
    )

    if not uecm_df.empty:
        df["consumption"] = df.apply(
            lambda x: x["rt_cons"] if pd.isnull(x["uecm"]) else x["uecm"], axis=1
        )
    else:
        df["consumption"] = df["rt_cons"].copy()

    if include_contract_symbol:
        try:
            df["contract"] = df["dt"].apply(lambda x: iso_to_contract(x))
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    return df



from eptr2 import EPTR2
import pandas as pd
from eptr2.util.time import iso_to_contract, get_utc3_now
import time
from datetime import datetime, timedelta


def get_bpm_range(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    max_lives: int = 2,
    verbose: bool = False,
    strict: bool = True,
    include_contract_symbol: bool = True,
) -> pd.DataFrame:
    """
    This function retrieves BPM (Balancing Power Market) weighted average data for multiple days between the specified start and end dates.

    start_date: str
        The start date in "YYYY-MM-DD" format.
    end_date: str
        The end date in "YYYY-MM-DD" format.
    max_lives: int
        The maximum number of retries for fetching data in case of errors.
    verbose: bool
        If True, prints detailed information about the fetching process.
    strict: bool
        If True, raises an exception if data fetching fails after retries.
    include_contract_symbol: bool
        If True, adds a column with the contract symbol derived from the date.
    Returns:
    pd.DataFrame
        A DataFrame containing the BPM data with columns for date, time, and YAL-YAT values.
    """

    date_range = pd.date_range(start=start_date, end=end_date, freq="D").to_list()

    date_range = sorted([d.strftime("%Y-%m-%d") for d in date_range])

    main_df = pd.DataFrame()

    today = get_utc3_now().strftime("%Y-%m-%d")

    for i, date_str in enumerate(date_range):

        if date_str > today:
            print("Skipping future dates:", date_str)
            break

        if verbose:
            print(f"Fetching data for {date_str} ({i + 1}/{len(date_range)})...")
        lives = max_lives

        while lives > 0:
            try:
                df = eptr.call(
                    "bpm-orders-w-avg",
                    date=date_str,
                    request_kwargs={"timeout": 5},
                )
                if df.empty:
                    print(f"No data found for {date_str}. Skipping...")
                else:
                    df.drop(columns=["date"], inplace=True)
                    df.rename(columns={"time": "dt"}, inplace=True)
                    main_df = pd.concat([main_df, df])
                break
            except Exception as e:
                if lives > 0:
                    print(
                        f"Error fetching data for {date_str}: {e}. Retrying after sleeping..."
                    )
                    lives -= 1
                    time.sleep(max_lives - lives)  # Sleep longer with each retry
                else:
                    print(f"Failed to fetch data for {date_str} after retries.")
                    if strict:
                        raise Exception("Failed to fetch data after retries.")

    main_df.reset_index(drop=True, inplace=True)

    if include_contract_symbol:
        try:
            main_df["contract"] = main_df["dt"].apply(lambda x: iso_to_contract(x))
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    return main_df


def get_bpm_period(
    eptr: EPTR2,
    period: str,
    max_lives: int = 2,
    verbose: bool = False,
    strict: bool = True,
    include_contract_symbol: bool = True,
) -> pd.DataFrame:
    """
    This function is a wrapper for `get_bpm_range` that retrieves BPM data for a specific month, regardless of the day.

    eptr: EPTR2
        An instance of the EPTR2 class to interact with the EPT API.
    period: str
        The period for which to retrieve BPM data, formatted as "YYYY-MM-DD".
    max_lives: int
        The maximum number of retries for fetching data in case of errors.
    verbose: bool
        If True, prints detailed information about the fetching process.
    strict: bool
        If True, raises an exception if data fetching fails after retries.
    include_contract_symbol: bool
        If True, adds a column with the contract symbol derived from the date.
    Returns:
    pd.DataFrame
        A DataFrame containing the BPM data for the specified month, with columns for date, time, and YAL-YAT values.
    """

    start_dt = datetime.strptime(period, "%Y-%m-%d").replace(day=1)
    end_dt = (start_dt + timedelta(days=31)).replace(day=1) - timedelta(days=1)
    start_date = start_dt.strftime("%Y-%m-%d")
    end_date = end_dt.strftime("%Y-%m-%d")

    df = get_bpm_range(
        eptr=eptr,
        start_date=start_date,
        end_date=end_date,
        max_lives=max_lives,
        verbose=verbose,
        strict=strict,
        include_contract_symbol=include_contract_symbol,
    )

    return df



from datetime import datetime, timedelta, timezone
import pytz


def check_iso_format(
    val: str | datetime,
    strict_hour_format: bool = False,
    convert_to_hour_format: bool = False,
):
    """
    Check if a string is datetime in ISO format and convert if necessary.
    Hour formatted dt is like 2023-01-01T10:00:00+03:00 (no minute, second, microsecond, tz is UTC+3)
    """
    try:
        dt = datetime.fromisoformat(val) if isinstance(val, str) else val
        if convert_to_hour_format:
            dt = dt.replace(tzinfo=timezone(timedelta(seconds=10800)))
            dt = dt.replace(microsecond=0)
            dt = dt.replace(second=0)
            dt = dt.replace(minute=0)

            return dt
        elif strict_hour_format:
            if (
                dt.minute != 0
                or dt.second != 0
                or dt.microsecond != 0
                or dt.tzinfo == timezone(timedelta(seconds=10800))
            ):
                return None
        return dt
    except ValueError:
        return None
    except AttributeError:
        return None


def get_utc3_now():
    """
    Gets Turkey time (UTC +3)
    """
    return datetime.now(tz=pytz.timezone("Europe/Istanbul"))


def get_today_utc3():
    """
    Gets Turkey today (UTC +3)
    """
    return get_utc3_now().strftime("%Y-%m-%d")


def get_date_from_key(key):
    if key == "today":
        res = get_today_utc3()
    elif key == "start_of_month":
        res = get_utc3_now().strftime("%Y-%m-01")

    return res


def parse_dt_transparency_1(dt):
    ## Example 2023-07-01T00:00:00.000+0300
    return datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S.%f%z")


def datetime_to_contract(
    dt, is_ts=False, root_only=False, block_hours: int | None = None
):
    if dt is None:
        return None

    if isinstance(dt, str):
        dt = check_iso_format(val=dt)
        if dt is None:
            dt = datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")

    if is_ts:
        dt = datetime.utcfromtimestamp(dt) + timedelta(hours=3)
    c_raw = dt.strftime("%y%m%d%H")
    if root_only:
        return c_raw
    elif block_hours is None:
        return "PH" + c_raw
    else:
        if block_hours > 24 or block_hours < 1:
            raise ValueError("block_hours must be between 1 and 24")

        return "PB" + c_raw + "-" + str(100 + block_hours)[1:]


def iso_to_contract(dt):
    return datetime_to_contract(check_iso_format(dt, convert_to_hour_format=True))


def contract_to_datetime(
    contract, timestamp: bool = False, localize: bool = True, to_str: bool = False
):
    dt_obj = datetime.strptime(
        contract[2:10] + ":00:00" + ("+03:00" if localize else ""), "%y%m%d%H:%M:%S%z"
    )

    if to_str:
        return dt_obj.isoformat(timespec="seconds")

    if timestamp:
        dt_obj = dt_obj.timestamp()
    return dt_obj


def get_hourly_date_range(start_date, end_date, return_str=False):
    l = []
    start_date = datetime.strptime(start_date, "%Y-%m-%d")
    end_date = (
        datetime.strptime(end_date, "%Y-%m-%d") + timedelta(days=1) - timedelta(hours=1)
    )
    t = 0
    while True:
        l.append(start_date + timedelta(hours=t))
        if l[-1] >= end_date:
            break
        t += 1

    if return_str:
        l = [str(x) for x in l]

    return l


def get_hourly_contract_range_list(start_date, end_date):
    l = [
        datetime_to_contract(x)
        for x in get_hourly_date_range(start_date, end_date, return_str=False)
    ]

    return l


def contract_to_hour(c):
    """
    Get the hour from a contract with the format PHyyMMDDHH or PByyMMDDHH-BB
    """
    return c[8:10]


def contract_to_day(c):
    """
    Get the day from a contract with the format PHyyMMDDHH or PByyMMDDHH-BB
    """
    return c[6:8]


def contract_to_wday(c):
    """
    Get the weekday from a contract with the format PHyyMMDDHH or PByyMMDDHH-BB
    """
    return contract_to_datetime(c).strftime("%w")


def contract_to_date_info(c, include_c=False):
    c_ts = contract_to_datetime(c, timestamp=True)
    d = {
        "wday": ts_to_day_of_week_utc3(c_ts),
        "hour": ts_to_hour_utc3(c_ts),
        "day": ts_to_day_utc3(c_ts),
    }

    if include_c:
        d["c"] = c

    return d


def contract_to_day_ahead_time(
    c: str, da_hour: int = 11, da_minutes: int = 0, as_ts=True
):
    """
    Get the day ahead decision time from a contract with the format PHyyMMDDHH or PByyMMDDHH-BB
    """
    c_dt = contract_to_datetime(c, timestamp=False)
    da_dt = c_dt.replace(
        hour=da_hour, minute=da_minutes, second=0, microsecond=0
    ) - timedelta(days=1)

    if as_ts:
        da_ts = da_dt.timestamp()
        return da_ts

    return da_dt


def ts_to_day_of_week_utc3(ts):
    """
    0: Sunday
    """
    return ts_to_format_utc3(ts=ts, format="%w")


def ts_to_format_utc3(ts, format="%Y-%m-%d"):
    return (datetime.fromtimestamp(ts, tz=pytz.timezone("Europe/Istanbul"))).strftime(
        format
    )


def ts_to_date_utc3(ts):
    """
    Convert timestamp to UTC3 - date
    """
    return ts_to_format_utc3(ts, format="%Y-%m-%d")


def ts_to_day_utc3(ts):
    """
    Convert timestamp to UTC3 and get the day
    """
    return ts_to_format_utc3(ts, format="%d")


def ts_to_hour_utc3(ts):
    """
    Convert timestamp to UTC3 and get the hour
    """
    return ts_to_format_utc3(ts, format="%H")


def calculate_active_contracts():

    now_dt = get_utc3_now()
    sod = now_dt.replace(minute=0, second=0, microsecond=0) + timedelta(hours=2)
    eod = now_dt.replace(hour=23, minute=0, second=0, microsecond=0)
    if now_dt.hour >= 18:
        eod += timedelta(days=1)

    l = [
        datetime_to_contract(sod + timedelta(hours=i))
        for i in range(int((eod - sod).total_seconds() // 3600) + 1)
    ]

    return l


def contract_close_time(c, to_timestamp=False, delta=3600):
    """
    Get the close time of a contract with the format PHyyMMDDHH or PByyMMDDHH-BB
    """
    dt = contract_to_datetime(c) - timedelta(seconds=delta)

    if to_timestamp:
        ts = dt.timestamp()
        return ts

    return dt


def time_to_contract_close(c, dt_then: datetime | str = None, ts_then: float = None):
    """
    Get the time to close of a contract with the format PHyyMMDDHH or PByyMMDDHH-BB
    """
    dt_close = contract_close_time(c, to_timestamp=True)
    if ts_then is not None:
        now_ts = ts_then
    elif dt_then is None:
        now_ts = get_utc3_now().timestamp()
    elif isinstance(dt_then, datetime):
        now_ts = dt_then.timestamp()
    elif isinstance(dt_then, str):
        try:
            now_ts = datetime.fromisoformat(dt_then).timestamp()
        except ValueError:
            print("Invalid datetime string format. ISO format expected.")
            return None

    return dt_close - now_ts


def get_time_min_max_price_map():

    map_l = [
        {
            "date": "2025-04-05",
            "min": 0,
            "max": 3400.0,
            "idm_min": 0,
            "idm_max": 3502.0,
        },
        {
            "date": "2024-06-01",
            "min": 0,
            "max": 3000.0,
            "idm_min": 0,
            "idm_max": 3090.0,
        },
        {
            "date": "2023-07-04",
            "min": 0,
            "max": 2700.0,
            "idm_min": 0,
            "idm_max": 2781.0,
        },
        {
            "date": "2023-04-01",
            "min": 0,
            "max": 2600.0,
            "idm_min": 0,
            "idm_max": 2678.0,
        },
        {
            "date": "2023-03-01",
            "min": 0,
            "max": 3050.0,
            "idm_min": 0,
            "idm_max": 3141.5,
        },
        {
            "date": "2023-02-01",
            "min": 0,
            "max": 3650.0,
            "idm_min": 0,
            "idm_max": 3759.5,
        },
        {
            "date": "2023-01-01",
            "min": 0,
            "max": 4200.0,
            "idm_min": 0,
            "idm_max": 4326.0,
        },
        {
            "date": "2022-09-02",
            "min": 0,
            "max": 4800.0,
            "idm_min": 0,
            "idm_max": 4944.0,
        },
    ]

    ## Sort the map by date

    map_l = sorted(map_l, key=lambda x: x["date"], reverse=True)

    return map_l


def contract_to_floor_ceil_prices(c):
    """
    Given a contract return the min and max prices
    """

    the_date = contract_to_datetime(c).date()
    mm_map = get_time_min_max_price_map()
    for x in mm_map:
        if the_date >= datetime.strptime(x["date"], "%Y-%m-%d").date():
            return x

    return None


def get_probable_settlement_date(x: str | None = None, settlement_day=15):
    """
    This function calculates the next probable settlement date based on certain criteria.
    """

    now_dt = get_utc3_now()

    if x is None:
        x = now_dt

    elif isinstance(x, str):
        x: datetime = datetime.strptime(x, "%Y-%m-%d")

    if x.month == 12:
        # If the month is December, we set the next month to January of the next year
        x = x.replace(year=x.year + 1, month=1)
    else:
        # Otherwise, we just increment the month by 1
        x = x.replace(month=x.month + 1)

    probable_settlement_date = x.replace(
        day=settlement_day,
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
    )

    return probable_settlement_date


def check_date_for_settlement(x: str | datetime, settlement_day=15):
    """
    Check if the settlement date for a given date has already occured.
    """

    now_dt = get_utc3_now()
    probable_settlement_date = get_probable_settlement_date(
        x=x, settlement_day=settlement_day
    )

    if now_dt.timestamp() >= probable_settlement_date.timestamp():
        return True

    return False



import os
from eptr2.tutorials.main import run_app
from eptr2.util.costs import calculate_imb_cost, calculate_diff_cost
from eptr2.composite.price_and_cost import get_hourly_price_and_cost_data


def run_calc_app(username: str, password: str, port: int | None = None):
    script_path = os.path.abspath(__file__)
    run_app(username=username, password=password, script_path=script_path, port=port)


def map_source(source: str):
    if source == "wind":
        return "💨 Rüzgar"
    elif source == "solar":
        return "☀️ Güneş"
    elif source == "other":
        return "Diğer"


if __name__ == "__main__":

    import streamlit as st
    from eptr2 import EPTR2
    from datetime import datetime, timedelta
    import pytz
    import pandas as pd

    username = os.getenv("EPTR_USERNAME")
    password = os.getenv("EPTR_PASSWORD")

    st.set_page_config(
        page_title="Dengesizlik/Imbalance",
        page_icon="👩‍💻",
        layout="wide",
        initial_sidebar_state="auto",
        menu_items=None,
    )

    ss = st.session_state
    ss["lang"] = ss.get("lang", "tr")
    ss["eptr"] = ss.get(
        "eptr",
        EPTR2(
            username=username,
            password=password,
            postprocess=True,
        ),
    )

    ss["cdf"] = ss.get("cdf", None)

    ss["stats"] = ss.get("stats", None)
    ss["gen_type"] = ss.get("gen_type", None)

    def get_data():
        refresh_data = True
        if ss["cdf"] is not None:
            print("cdf is not None")
            if ss["date"] == ss["cdf"]["date"][0].date():
                refresh_data = False

        if refresh_data:
            print("Getting data...")
            ss["cdf"] = get_hourly_price_and_cost_data(
                eptr=ss["eptr"],
                start_date=str(ss["date"]),
                end_date=str(ss["date"]),
                include_wap=True,
                add_kupst_cost=True,
                verbose=False,
            )
            # ss["cdf"] = ss["eptr"].call(
            #     "mcp-smp-imb", start_date=str(ss["date"]), end_date=str(ss["date"])
            # )
            ss["cdf"]["date"] = pd.to_datetime(ss["cdf"]["date"])

        ss["stats"] = dict(ss["cdf"][ss["cdf"]["date"].dt.hour == ss["hour"]].iloc[0])

        print("==*****==")
        print(ss["org_type"])
        print(ss["gen_type"])
        print("==*****==")

        ss["stats"]["costs"] = calculate_diff_cost(
            actual=ss["actual"],
            forecast=ss["forecast"],
            prod_source=ss.get("gen_type", None),
            mcp=ss["stats"]["mcp"],
            smp=ss["stats"]["smp"],
            dsg_absorption_rate=1.0,
            return_detail=True,
        )

    st.sidebar.subheader("Parametreler")

    st.sidebar.date_input(
        "Tarih",
        value=datetime.now(tz=pytz.timezone("Europe/Istanbul")) - timedelta(days=1),
        key="date",
    )

    st.sidebar.selectbox(
        "Saat", list(range(24)), format_func=lambda x: f"{x:02d}:00", key="hour"
    )

    st.sidebar.radio(
        "Organizasyon Türü", options=["Tüketici", "Üretici"], key="org_type"
    )

    if ss["org_type"] == "Üretici":
        st.sidebar.radio(
            "Santral Tipi",
            options=["wind", "solar", "other"],
            format_func=lambda x: map_source(x),
            index=2,
            key="gen_type",
        )

    st.sidebar.number_input(
        "Tahmin (MWh)", min_value=0.0, max_value=5000.0, step=0.1, key="forecast"
    )
    st.sidebar.number_input(
        "Gerçekleşme (MWh)", min_value=0.0, max_value=5000.0, step=0.1, key="actual"
    )

    st.sidebar.button("💰 Hesapla", key="calc", on_click=get_data, type="primary")

    st.sidebar.caption(
        "Bu uygulama eptr2 Python kütüphanesinden çalıştırılmaktadır. Fiyat ve maliyetlerde EPİAŞ Şeffaflık Platformu verileri kullanılmaktadır. Bütün hesap ve işlemler bilgilendirme ve eğitimr amaçlıdır. Tideseed/Robokami bilgilerin doğruluğu ile ilgili ve/veya kullanımdan dolayı herhangi bir sorumluluk kabul etmemektedir. Detaylı bilgi için [iletişim formu](https://robokami.com/#iletisim)nu kullanabilirsiniz."
    )

    st.header("eptr2: Fiyat ve Maliyet Hesaplama Eğitimi")

    with st.expander("Fiyat ve Maliyet Tablosu", expanded=False):
        if ss["cdf"] is not None:
            st.dataframe(ss["cdf"])
        else:
            st.warning(
                "Henüz fiyat ve maliyet verileri çekilmedi. Bir tarih seçip Hesapla tuşuna basın."
            )

    # with st.expander("Debug", expanded=False):

    #     st.json(ss)

    if ss["stats"]:
        st.caption(
            "⚡️ Miktar (MWh), 💰 Fiyat (TL/MWh), 💸 Maliyet (TL), 💸💸 Toplam Maliyet (TL)"
        )
        cols = st.columns([3, 3, 3, 3])

        with cols[0]:
            st.metric(
                "PTF 💰", round(ss["stats"]["mcp"], 2), help="Piyasa Takas Fiyatı"
            )
            st.metric(
                "SMF 💰", round(ss["stats"]["smp"], 2), help="Sistem Marjinal Fiyatı"
            )
            st.metric(
                "PDF 💰",
                round(ss["stats"]["pos_imb_price"], 2),
                help="Pozitif Dengesizlik Fiyatı: min(PTF, SMF)*0.97",
            )
            st.metric(
                "NDF 💰",
                round(ss["stats"]["neg_imb_price"], 2),
                help="Negatif Dengesizlik Fiyatı: max(PTF, SMF)*1.03",
            )

        with cols[1]:
            st.metric(
                "PDM 💸",
                round(ss["stats"]["pos_imb_cost"], 2),
                help="Pozitif Dengesizlik Maliyeti: PTF - min(PTF, SMF)*0.97",
            )
            st.metric(
                "NDM 💸",
                round(ss["stats"]["neg_imb_cost"], 2),
                help="Negatif Dengesizlik Maliyeti: max(PTF, SMF)*1.03 - PTF",
            )
            st.metric(
                "Birim KÜPST 💸",
                round(ss["stats"]["kupst_cost"], 2),
                help="Birim KÜPST maliyeti max(PTF,SMF) * 0.03",
            )
            st.metric(
                "Sistem Yönü 🔛",
                ss["stats"]["sd_sign"],
                help="1: Enerji Fazlası, -1: Enerji Açığı, 0: Dengede",
            )

        with cols[2]:
            st.metric(
                "Brüt Dengesizlik ⚡️",
                ss["stats"]["costs"]["imbalances"]["diff"],
                help="Eksi değer negatif dengesizlik, artı deger pozitif dengesizlik demektir.",
            )
            st.metric(
                "DSG Tolerans ⚡️",
                ss["stats"]["costs"]["imbalances"]["imb_tol_value"],
                help="DSG Tolerans Miktarı (Tolerans Oranı: 10%)",
            )
            st.metric(
                "Bireysel Dengesizlik ⚡️",
                ss["stats"]["costs"]["imbalances"]["individual_imbalance"],
                help="DSG toleransı üzerinde kalan dengesizlik miktarı.",
            )

            if ss["org_type"] == "Üretici" and "kupst" in ss["stats"]["costs"].keys():
                st.metric(
                    "KÜPSM Tolerans ⚡️",
                    round(ss["stats"]["costs"]["kupst"]["kupsm_tol"], 2),
                    help="KÜPSM toleransı. Kaynak "
                    + map_source(ss["gen_type"])
                    + "("
                    + format(ss["stats"]["costs"]["kupst"]["kupst_tol_perc"], ".2f")
                    + ")",
                )
                st.metric(
                    "KÜPSM ⚡️",
                    round(ss["stats"]["costs"]["kupst"]["kupsm"], 2),
                    help="KÜPSM",
                )

        with cols[3]:
            ind_cost = round(ss["stats"]["costs"]["costs"]["ind_imbalance_cost"], 2)
            st.metric(
                "Bireysel Dengesizlik Maliyeti 💸",
                ind_cost,
                help="DSG toleransının üzerinde kalan dengesizlik maliyeti.",
            )

            dsg_cost = round(ss["stats"]["costs"]["costs"]["net_dsg_imb_cost"], 2)
            st.metric(
                "DSG İçi Dengesizlik Maliyeti 💸",
                dsg_cost,
                help="DSG toleransında kalıp absorbe edilemeyen dengesizlik maliyeti.",
            )

            if ss["org_type"] == "Üretici" and "kupst" in ss["stats"]["costs"].keys():
                kupst_cost = round(ss["stats"]["costs"]["kupst"]["kupst_cost"], 2)
                st.metric(
                    "KÜPST 💸",
                    kupst_cost,
                    help="KÜPSM",
                )
            else:
                kupst_cost = 0

            total_cost = ind_cost + dsg_cost + kupst_cost

            st.metric(
                "Toplam Maliyet 💸💸",
                f"{total_cost} TL",
                help="Dengesizlik maliyetleri + KÜPST maliyeti.",
            )

            unit_cost = round(total_cost / max(1, ss["actual"]))

            st.metric(
                "Birim Maliyet 💸⚖️",
                f"{unit_cost} TL/MWh",
                help="Birim dengesizlik maliyetleri + KÜPST maliyeti.",
            )
    else:
        st.info("Hesapla tuşuna basarak değerleri görüntüleyebilirsiniz.")
    # if ss["cdf"] is not None:
    #     st.dataframe(ss["cdf"])



from eptr2.composite.consumption import *
from eptr2.composite.price_and_cost import *
from eptr2.composite.production import *
from eptr2.composite.dabi import *
from eptr2.composite.bpm import *
from eptr2.composite.idm_log import *
from eptr2.composite.mms import *



import os
import subprocess


def run_app(username: str, password: str, script_path: str, port: int | None = None):
    """
    Launch the Streamlit app programmatically.
    """
    # Get the absolute path of this script
    # script_path = os.path.abspath(__file__)

    os.environ["EPTR_USERNAME"] = username
    os.environ["EPTR_PASSWORD"] = password

    # Run Streamlit CLI command to start the app
    run_l = ["streamlit", "run", script_path]
    if port is not None:
        run_l.extend(["--server.port", str(port)])
    try:
        subprocess.run(run_l, check=True)
    except FileNotFoundError:
        raise RuntimeError(
            "Streamlit is not installed or not found in the PATH. Please ensure Streamlit is installed in your environment."
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to run the Streamlit app: {e}")



from eptr2 import EPTR2
from eptr2.util.time import iso_to_contract
import pandas as pd


def get_mms_detail(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    org_id: str | None = None,
    uevcb_id: str | None = None,
    pp_id: str | None = None,
    message_type_id: str | None = None,
    include_contract_symbol: bool = False,
    include_summary: bool = False,
    verbose: bool = False,
):
    """
    This composite function gets market message system (MMS) data, expands it and adds the contract symbols (optional).
    """

    df: pd.DataFrame = eptr.call(
        "mms",
        start_date=start_date,
        end_date=end_date,
        org_id=org_id,
        uevcb_id=uevcb_id,
        pp_id=pp_id,
        message_type_id=message_type_id,
    )

    df_wip = df.explode("faultDetails", ignore_index=True)

    df_res = df_wip.drop(columns=["faultDetails"]).join(
        pd.json_normalize(df_wip["faultDetails"], sep="_")
    )

    if include_summary and not include_contract_symbol:
        if verbose:
            print(
                "Contract symbols are required for summary. Setting include_contract_symbol to True."
            )
        include_contract_symbol = True

    if include_contract_symbol:
        df_res["contract"] = df_res["hour"].apply(iso_to_contract)

    if include_summary:
        if verbose:
            print("Getting MMS loss summary by contract...")
        df_summary = get_mms_loss_summary_by_contract(df_res)
        return {"detail": df_res, "summary": df_summary}

    return df_res


def get_mms_loss_summary_by_contract(df: pd.DataFrame):
    """
    This function gets the loss summary of the MMS data.
    """

    summary_df = (
        df.groupby("contract")[["faultCausedPowerLoss", "faultCausedEnergyLoss"]]
        .sum()
        .reset_index()
        .sort_values("contract", ignore_index=True)
    )

    return summary_df



def get_kupst_tolerance(source: str):
    alias_map = {"sun": "solar"}

    tol_source_map = {"wind": 0.17, "solar": 0.10}
    return tol_source_map.get(alias_map.get(source, source), 0.05)


def calculate_unit_imbalance_price(
    mcp: float, smp: float, penalty_margin: float = 0.03
):
    d = {
        "pos": (1 - penalty_margin) * min(mcp, smp),
        "neg": (1 + penalty_margin) * max(mcp, smp),
    }
    return d


def calculate_unit_imbalance_cost(mcp: float, smp: float, penalty_margin: float = 0.03):
    d = calculate_unit_imbalance_price(mcp=mcp, smp=smp, penalty_margin=penalty_margin)

    d["pos"] = mcp - d["pos"]
    d["neg"] = d["neg"] - mcp

    return d


def calculate_unit_kupst_cost(mcp: float, smp: float, kupst_multiplier: float = 0.03):
    return max(mcp, smp) * kupst_multiplier


def calculate_imbalance_amounts(
    actual: float,
    forecast: float,
    is_producer: bool,
    imb_tol: float = 0.1,
    dsg_absorption_rate: float = 0.0,
):
    """
    This function returns the imbalance amounts given actual and forecast values, imbalance tolerance and DSG absorption rate. DSG (Dengeden Sorumlu Grup / Balancing Responsible Group) can only absorb imbalances by the imb_tol proportion (default 10% as the regulatory standard). dsg_absorption_rate is the rate at which DSG absorbs imbalances. Anything above 10% imbalance is recorded as individual imbalance.

    For instance if actual is 100, forecast is 50 (current position) and dsg_absorption_rate is 0.5; DSG tolerance is 100*0.1 = 10Mwh (as gross imbalance is 50), DSG absorbed imbalance is 10*0.5 = 5Mwh, individual imbalance is 50-10 = 40 and total penalized imbalance is 5+40 = 45Mwh.
    """
    diff = forecast - actual

    if is_producer:
        diff = -diff

    imb_tol_value = round(actual * imb_tol, 1)

    sign = -1 if diff < 0 else 1

    dsg_imbalance = min(abs(diff), imb_tol_value) * sign
    net_dsg_imbalance = dsg_imbalance * (1 - dsg_absorption_rate)

    individual_imbalance = diff - dsg_imbalance

    res_d = {
        "diff": diff,
        "dsg_imbalance": dsg_imbalance,
        "net_dsg_imbalance": net_dsg_imbalance,
        "individual_imbalance": individual_imbalance,
        "imb_tol_value": imb_tol_value,
    }

    return res_d


def calculate_imb_cost(
    actual: float,
    forecast: float,
    mcp: float,
    smp: float,
    is_producer: bool,
    imb_tol: float = 0.1,
    return_detail: bool = False,
    dsg_absorption_rate: float = 0.0,
):

    ## Get the net costs of per imbalance MWh
    unit_cost_d = calculate_unit_imbalance_cost(mcp=mcp, smp=smp)

    ## Calculate the imbalance amounts
    res_d = calculate_imbalance_amounts(
        actual=actual,
        forecast=forecast,
        is_producer=is_producer,
        imb_tol=imb_tol,
        dsg_absorption_rate=dsg_absorption_rate,
    )

    dsg_imbalance = res_d["dsg_imbalance"]
    net_dsg_imbalance = res_d["net_dsg_imbalance"]
    individual_imbalance = res_d["individual_imbalance"]

    imb_side = "neg" if individual_imbalance < 0 else "pos"
    unit_cost = unit_cost_d[imb_side]

    dsg_imb_cost = abs(dsg_imbalance) * unit_cost
    net_dsg_imb_cost = abs(net_dsg_imbalance) * unit_cost
    individual_imb_cost = abs(individual_imbalance) * unit_cost
    total_imb_cost = individual_imb_cost + net_dsg_imb_cost

    if return_detail:
        return {
            "imbalances": res_d,
            "imb_side": imb_side,
            "costs": {
                "ind_imbalance_cost": individual_imb_cost,
                "dsg_imb_cost": dsg_imb_cost,
                "net_dsg_imb_cost": net_dsg_imb_cost,
                "total_imb_cost": total_imb_cost,
            },
        }
    else:
        return total_imb_cost


def calculate_imbalance_cost_deprecated(
    actual: float,
    forecast: float,
    mcp: float,
    smp: float,
    is_producer: bool,
    imbalance_discount: float = 0.0,
):
    if imbalance_discount > 1 or imbalance_discount < 0:
        raise ValueError("Imbalance dampening rate must be between 0 and 1")

    unit_cost_d = calculate_unit_imbalance_cost(mcp=mcp, smp=smp)

    diff = forecast - actual
    if is_producer:
        cost = abs(diff) * unit_cost_d["neg" if diff > 0 else "pos"]
    else:
        cost = abs(diff) * unit_cost_d["pos" if diff > 0 else "neg"]

    return cost * (1 - imbalance_discount)


def calculate_kupst_cost(
    actual: float,
    forecast: float,
    mcp: float,
    smp: float,
    tol: float | None = None,
    source: str | None = None,
    kupst_multiplier: float = 0.03,
    return_detail: bool = False,
):
    if tol is None:
        if source is None:
            raise Exception("Either tol(erance) or source parameter must be provided")
        tol = get_kupst_tolerance(source)

    tol_val = forecast * tol
    kupsm = max(0, abs(actual - forecast) - tol_val)
    unit_kupst = calculate_unit_kupst_cost(mcp, smp, kupst_multiplier)
    kupst_cost = kupsm * unit_kupst
    if return_detail:
        return {
            "kupst_tol_perc": tol,
            "kupsm_tol": tol_val,
            "kupsm": kupsm,
            "unit_kupst_cost": unit_kupst,
            "kupst_cost": kupst_cost,
        }

    return kupst_cost


def calculate_diff_cost(
    actual: float,
    forecast: float,
    mcp: float,
    smp: float,
    prod_source: str | None = None,
    imb_tol: float = 0.1,
    **kwargs
):
    is_producer = prod_source is not None

    res_d = calculate_imb_cost(
        actual=actual,
        forecast=forecast,
        mcp=mcp,
        smp=smp,
        is_producer=is_producer,
        imb_tol=imb_tol,
        dsg_absorption_rate=kwargs.get("dsg_absorption_rate", 0.0),
        return_detail=kwargs.get("return_detail", True),
    )

    if not kwargs.get("return_detail", True):
        res_d = {"total_imb_cost": res_d}

    if is_producer:
        kupst_d = calculate_kupst_cost(
            actual=actual,
            forecast=forecast,
            mcp=mcp,
            smp=smp,
            source=prod_source,
            tol=kwargs.get("kupst_tol", None),
            kupst_multiplier=kwargs.get("kupst_multiplier", 0.03),
            return_detail=kwargs.get("return_detail", True),
        )

        res_d["kupst"] = kupst_d

    return res_d


def calculate_diff_cost_deprecated(
    actual: float,
    forecast: float,
    mcp: float,
    smp: float,
    tol: float | None = None,
    kupst_multiplier: float = 0.03,
    is_producer: bool = True,
    source: str | None = None,
    imbalance_discount: float = 0.0,
):
    imbalance = calculate_imbalance_cost_deprecated(
        actual=actual,
        forecast=forecast,
        mcp=mcp,
        smp=smp,
        is_producer=is_producer,
        imbalance_discount=imbalance_discount,
    )

    kupst = calculate_kupst_cost(
        actual=actual,
        forecast=forecast,
        mcp=mcp,
        smp=smp,
        tol=tol,
        source=source,
        kupst_multiplier=kupst_multiplier,
    )

    total = imbalance + kupst

    return {"imbalance": imbalance, "kupst": kupst, "total": total}


def calculate_imbalance_prices_list(mcp: list, smp: list):
    """
    Calculates imbalance prices for positive and negative imbalances
    """
    d = {
        "pos": [min(x, y) * 0.97 for x, y in zip(mcp, smp)],
        "neg": [1.03 * max(x, y) for x, y in zip(mcp, smp)],
    }

    return d


def calculate_imbalance_cost_values_list(mcp: list, smp: list):
    """
    Calculates imbalance costs relative to day-ahead prices (MCP) for positive and negative imbalances
    """
    d = calculate_imbalance_prices_list(mcp, smp)
    d["pos"] = [x - y for x, y in zip(mcp, d["pos"])]
    d["neg"] = [y - x for x, y in zip(mcp, d["neg"])]

    return d


def calculate_kupst_cost_list(
    actual_values: list,
    forecast_values: list,
    mcp: list,
    smp: list,
    tol: float | None = None,
    source: str | None = None,
    kupst_multiplier: float = 0.03,
    min_kupst: float = 750.0,
):
    """
    Calculates production plan difference (KUPST) costs.
    """

    if tol is None:
        if source is None:
            raise Exception("Either tol(erance) or source parameter must be provided")
        tol = get_kupst_tolerance(source)

    kupst_cost_list = []

    for x, y, m, s in zip(forecast_values, actual_values, mcp, smp):
        cost = max(abs(x - y) - tol * x, 0) * max(m, s) * kupst_multiplier
        cost = max(cost, min_kupst)

        kupst_cost_list.append(cost)

    return kupst_cost_list


def calculate_imbalance_cost_list(
    actual_values: list,
    forecast_values: list,
    cost_d: dict,
    is_producer: bool,
    imbalance_discount: float = 0.0,
):
    """
    Calculates imbalance costs for a given forecast and actual values.
    Imbalance dampening reduces imbalance costs by the given rate (<1). It is used to emulate balance responsible party (DSG) effect.
    """

    if imbalance_discount > 1 or imbalance_discount < 0:
        raise ValueError("Imbalance dampening rate must be between 0 and 1")

    if len(actual_values) != len(forecast_values):
        raise ValueError("Actual and forecast values must have the same length")

    imbalance_cost_list = []

    for x, y, n, p in zip(forecast_values, actual_values, cost_d["neg"], cost_d["pos"]):
        if x < y:
            cost = (y - x) * p if is_producer else n
        else:
            cost = (x - y) * n if is_producer else p

        imbalance_cost_list.append(cost * (1 - imbalance_discount))

    return imbalance_cost_list


def calculate_diff_costs_list(
    actual_values: list,
    forecast_values: list,
    mcp: list,
    smp: list,
    is_producer: bool,
    tol: float | None = None,
    source: str | None = None,
    kupst_multiplier: float = 0.03,
    imbalance_discount: float = 0,
    min_kupst: float = 0.0,
):
    """
    Wrapper function to calculate all costs related to the difference between actual and forecasted values..
    """
    res_d = {"actual": actual_values, "forecast": forecast_values}

    cost_d = calculate_imbalance_cost_values_list(mcp, smp)
    res_d["diff"] = [x - y for x, y in zip(forecast_values, actual_values)]

    if is_producer:
        res_d["imbalance_direction"] = [
            "pos" if x < 0 else "neg" for x in res_d["diff"]
        ]
    else:
        res_d["imbalance_direction"] = [
            "pos" if x > 0 else "neg" for x in res_d["diff"]
        ]

    res_d["imbalance_cost"] = calculate_imbalance_cost_list(
        actual_values=actual_values,
        forecast_values=forecast_values,
        cost_d=cost_d,
        is_producer=is_producer,
        imbalance_discount=imbalance_discount,
    )

    if is_producer:
        res_d["kupst_cost"] = calculate_kupst_cost_list(
            actual_values=actual_values,
            forecast_values=forecast_values,
            mcp=mcp,
            smp=smp,
            tol=tol,
            source=source,
            kupst_multiplier=kupst_multiplier,
            min_kupst=min_kupst,
        )

    return res_d



def process_special_calls(call_key, param_d):

    if call_key in ["bpm-orders", "bpm-orders-w-avg"]:
        if "date_time" in param_d.keys():
            raise Exception(
                f"date_time parameter is not supported for {call_key}. Use 'date' instead."
            )

    if call_key in [
        "ng-vgp-contract-price-summary-period",
        "ng-vgp-ggf-period",
        "ng-vgp-matched-quantity-period",
        "ng-vgp-open-positions-period",
        "ng-vgp-transaction-history-period",
        "ng-vgp-transaction-volumes-period",
    ]:
        param_d["is_txn_period"] = False
    elif call_key in [
        "ng-vgp-contract-price-summary-se",
        "ng-vgp-ggf-se",
        "ng-vgp-matched-quantity-se",
        "ng-vgp-open-positions-se",
        "ng-vgp-transaction-history-se",
        "ng-vgp-transaction-volumes-se",
    ]:
        param_d["is_txn_period"] = True

    return param_d



from eptr2 import EPTR2
from eptr2.util.costs import calculate_unit_kupst_cost
from eptr2.util.time import iso_to_contract
import pandas as pd


def get_hourly_price_and_cost_data(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    include_wap: bool = True,
    add_kupst_cost: bool = True,
    verbose: bool = False,
    include_contract_symbol: bool = False,
    timeout: int = 10,
):
    """
    This composite function gets price and imbalance (kupst included) cost data.
    Imbalance cost data consists of MCP (PTF), SMP (SMF), system direction (Enerji Açığı - up regulated, Energy Fazlası - down regulated, Dengede - Balanced), positive imbalance cost and negative imbalance cost.

    Resulting columns are:
    - date: Datetime in ISO format and +03:00 timezone
    - mcp: Market Clearing Price (MCP/PTF) in TL (Turkish Lira)
    - smp: System Marginal Price (SMP/SMF) in TL (Turkish Lira)
    - wap: Weighted Average Price (WAP) in TL (Turkish Lira)
    - system_direction: System direction (Enerji Açığı - up regulated, Energy Fazlası - down regulated, Dengede - Balanced)
    - sd_sign: System direction sign (1 for Enerji Fazlası, -1 for Enerji Açığı, 0 for Dengede)
    - pos_imb_price: Positive imbalance price in TL (Turkish Lira)
    - neg_imb_price: Negative imbalance price in TL (Turkish Lira) (negative value)
    - pos_imb_cost: Positive imbalance cost in TL (Turkish Lira) (ptf - pos_imb_price)
    - neg_imb_cost: Negative imbalance cost in TL (Turkish Lira) (neg_imb_price - ptf)
    """

    if verbose:
        print("Getting MCP, SMP and imbalance price data...")

    price_df = eptr.call(
        "mcp-smp-imb",
        start_date=start_date,
        end_date=end_date,
        request_kwargs={"timeout": timeout},
    )

    price_df.drop(columns=["time"], inplace=True)

    price_df.rename(
        columns={
            "ptf": "mcp",
            "smf": "smp",
            "positiveImbalance": "pos_imb_price",
            "negativeImbalance": "neg_imb_price",
            "systemStatus": "system_direction",
        },
        inplace=True,
    )

    price_df["sd_sign"] = price_df["system_direction"].apply(
        lambda x: -1 if x == "Enerji Açığı" else 1 if x == "Enerji Fazlası" else 0
    )

    price_df["pos_imb_cost"] = price_df["mcp"] - price_df["pos_imb_price"]
    price_df["neg_imb_cost"] = price_df["neg_imb_price"] - price_df["mcp"]

    if include_wap:
        if verbose:
            print("Getting WAP data...")
        wap_df = eptr.call(
            "wap",
            start_date=start_date,
            end_date=end_date,
            request_kwargs={"timeout": timeout},
        )
        wap_df.drop(columns=["hour"], inplace=True)

        price_df = price_df.merge(wap_df, on="date", how="outer")

    if add_kupst_cost:
        if verbose:
            print("Calculating unit KUPST cost...")

        price_df["kupst_cost"] = price_df.apply(
            lambda x: calculate_unit_kupst_cost(mcp=x["mcp"], smp=x["smp"]), axis=1
        )

    if include_contract_symbol:
        try:
            price_df["contract"] = price_df["date"].apply(lambda x: iso_to_contract(x))
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    ### Column ordering
    columns_order = [
        "date",
        "contract",
        "pos_imb_vol",
        "neg_imb_vol",
        "pos_imb_mwh",
        "neg_imb_mwh",
        "mcp",
        "wap",
        "smp",
        "pos_imb_price",
        "neg_imb_price",
        "system_direction",
        "sd_sign",
        "pos_imb_cost",
        "neg_imb_cost",
        "kupst_cost",
    ]

    columns_order = [col for col in columns_order if col in price_df.columns]

    price_df = price_df[columns_order]

    return price_df


def get_hourly_imbalance_data(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    include_price_and_cost_data: bool = True,
    include_wap: bool = True,
    add_kupst_cost: bool = True,
    verbose: bool = False,
    include_contract_symbol=False,
    timeout: int = 10,
):
    """
    This composite function gets imbalance volume, imbalance quantity and imbalance cost data.
    Imbalance cost data consists of MCP (PTF), SMP (SMF), system direction (Enerji Açığı - up regulated, Energy Fazlası - down regulated, Dengede - Balanced), positive imbalance cost and negative imbalance cost.

    Resulting columns are:
    - date: Datetime in ISO format and +03:00 timezone
    - pos_imb_vol: Positive imbalance volume in TL (Turkish Lira)
    - neg_imb_vol: Negative imbalance volume in TL (Turkish Lira) (negative value)
    - pos_imb_mwh: Positive imbalance quantity in MWh
    - neg_imb_mwh: Negative imbalance quantity in MWh (negative value)
    - mcp: Market Clearing Price (MCP/PTF) in TL (Turkish Lira)
    - smp: System Marginal Price (SMP/SMF) in TL (Turkish Lira)
    - system_direction: System direction (Enerji Açığı - up regulated, Energy Fazlası - down regulated, Dengede - Balanced)
    - sd_sign: System direction sign (1 for Enerji Fazlası, -1 for Enerji Açığı, 0 for Dengede)
    - pos_imb_price: Positive imbalance price in TL (Turkish Lira)
    - neg_imb_price: Negative imbalance price in TL (Turkish Lira) (negative value)
    - pos_imb_cost: Positive imbalance cost in TL (Turkish Lira) (ptf - pos_imb_price)
    - neg_imb_cost: Negative imbalance cost in TL (Turkish Lira) (neg_imb_price - ptf)
    """

    imb_vol_df = eptr.call(
        "imb-vol",
        start_date=start_date,
        end_date=end_date,
        verbose=verbose,
        request_kwargs={"timeout": timeout},
    )

    if imb_vol_df.empty:
        raise ValueError(
            "Imbalance data returns empty. Settlement data is published after the 15th of the next month (e.g. May data is published at June 15 earliest) or in the first following working day normally. It is also possible for the imbalance data to be released later than settlement. Check the date range."
        )

    imb_vol_df.drop(columns=["hour"], inplace=True)
    imb_vol_df.rename(
        columns={
            "positiveImbalance": "pos_imb_vol",
            "negativeImbalance": "neg_imb_vol",
        },
        inplace=True,
    )

    imb_vol_df["neg_imb_vol"] = -imb_vol_df["neg_imb_vol"]

    ## TEMP FIX for API quirk (i.e. it does not return data unless start date is the first day of the month)
    start_date_temp = pd.to_datetime(start_date).replace(day=1).strftime("%Y-%m-%d")

    imb_qty_df = eptr.call(
        "imb-qty",
        start_date=start_date_temp,
        end_date=end_date,
        verbose=verbose,
        request_kwargs={"timeout": timeout},
    )

    imb_qty_df = imb_qty_df[imb_qty_df["date"] >= start_date]

    # imb_qty_df.drop(columns=["hour"], inplace=True)

    imb_qty_df.rename(
        columns={
            "positiveImbalance": "pos_imb_mwh",
            "negativeImbalance": "neg_imb_mwh",
        },
        inplace=True,
    )

    imb_qty_df["neg_imb_mwh"] = -imb_qty_df["neg_imb_mwh"]

    merged_df = imb_vol_df.merge(imb_qty_df, on="date", how="outer")

    if include_contract_symbol:
        try:
            merged_df["contract"] = merged_df["date"].apply(
                lambda x: iso_to_contract(x)
            )
        except Exception as e:
            print("Contract information could not be added. Error:", e)

    if not include_price_and_cost_data:
        return merged_df

    price_df = get_hourly_price_and_cost_data(
        eptr=eptr,
        start_date=start_date,
        end_date=end_date,
        include_wap=include_wap,
        add_kupst_cost=add_kupst_cost,
        verbose=verbose,
        include_contract_symbol=False,
        timeout=timeout,
    )

    merged_df = merged_df.merge(price_df, on="date", how="outer")

    return merged_df



from eptr2 import EPTR2
import pandas as pd
from datetime import datetime, timedelta
from eptr2.util.time import get_hourly_contract_range_list
import time


def idm_log_longer(
    eptr: EPTR2,
    start_date: str,
    end_date: str,
    contract_wise: bool = True,
    verbose: bool = False,
    trials: int = 3,
    cooldown: int = 15,
    days_interval: int = 6,
) -> pd.DataFrame:
    """
    This function gets the IDM log data for a longer period.

    If contract_wise is True, start_date is taken from a day earlier to get the previous day's transactions of the start_date's contracts. For instance if start_date is 2025-05-15, trading for PH25051500 starts at 2025-05-14 18:00:00. Therefore, the start_date is set to 2025-05-14. The end_date is not changed as contract closing times are before the contract times.
    """

    start_dt = datetime.strptime(start_date, "%Y-%m-%d")
    if contract_wise:
        ## Get conract range list
        c_l = get_hourly_contract_range_list(start_date=start_date, end_date=end_date)
        start_dt = start_dt - timedelta(days=1)
        start_date = start_dt.strftime("%Y-%m-%d")

    end_dt = datetime.strptime(end_date, "%Y-%m-%d")

    ## Make a sanity check and show a friendly warning
    whole_interval = (end_dt - start_dt).days
    if whole_interval > 32:
        print(
            "Friendly warning: Your time interval might be a bit long. You might lose all data if the API fails. We recommmend monthly aggregate calls in order to have proper checkpoints."
        )

    period_start_dt = start_dt
    period_end_dt = min(period_start_dt + timedelta(days=days_interval), end_dt)

    main_df = pd.DataFrame()
    while period_start_dt <= end_dt:

        try:
            period_start_date = period_start_dt.strftime("%Y-%m-%d")
            period_end_date = period_end_dt.strftime("%Y-%m-%d")
            if verbose:
                print(
                    f"Getting IDM log data for {period_start_date} to {period_end_date}"
                )

            # Get the IDM log data
            df = eptr.call(
                "idm-log",
                start_date=period_start_date,
                end_date=period_end_date,
            )

            main_df = pd.concat([main_df, df], ignore_index=True)

            period_start_dt = period_end_dt + timedelta(days=1)
            period_end_dt = min(period_start_dt + timedelta(days=days_interval), end_dt)
        except Exception as e:
            print(
                f"Error getting IDM log data for {period_start_date} to {period_end_date}"
            )
            print(e)

            trials -= 1

            if trials < 0:
                print("Max trials reached. Exiting.")
                break

            print(f"Retrying after {cooldown} seconds... {trials} trials left.")
            time.sleep(cooldown)
            continue

    if contract_wise:
        c_df = pd.DataFrame(data={"contractName": c_l})
        main_df = main_df.merge(c_df, on="contractName", how="inner")

    main_df = main_df.sort_values(by=["contractName", "date", "id"]).reset_index(
        drop=True
    )
    return main_df


def idm_log_period(eptr: EPTR2, period: str, **kwargs):
    """
    This function is a wrapper for the IDM log data for a specific monthly period (e.g. 2025-05-01).
    """
    ## Check if the period is a valid date
    try:
        period_dt = datetime.strptime(period, "%Y-%m-%d")
    except ValueError:
        raise ValueError("Invalid date format. Please use YYYY-MM-DD.")

    ## Start of month
    start_dt = period_dt.replace(day=1)
    ## End of month
    end_dt = (start_dt + timedelta(days=31)).replace(day=1) - timedelta(days=1)
    start_date = start_dt.strftime("%Y-%m-%d")
    end_date = end_dt.strftime("%Y-%m-%d")
    return idm_log_longer(
        eptr=eptr,
        start_date=start_date,
        end_date=end_date,
        **kwargs,
    )
